;---------------------------------------------------------------------------------------------
;
; LCD flat panel mode fix for X68000 by Neill Corlett
;
; The X68000 uses various 512-line screen modes which are incorrectly detected as 480-line on
; most LCD flat panels, causing parts of the screen to be cut off.  This program attempts to
; fix that by tweaking the CRTC timings to make them look more like 600-line.
;
; All timings are initialized from a configuration file (e.g. lcdfix.ini, included)
;
;---------------------------------------------------------------------------------------------

    .cpu 68000
    .text

peaw: .macro n
    .dc.w $4878,n
    .endm

;---------------------------------------------------------------------------------------------
;
; Beginning of resident portion
;
resident_start:

;---------------------------------------------------------------------------------------------
;
; Old IOCS $10 handler
;
old_iocs_10:
    .dc.w $4ef9 ; jmp.l
old_iocs_10_address:
    .dc.l 0

;
; New IOCS $10 handler
;
new_iocs_10:
    ;
    ; Are we checking the mode or setting it?
    ;
    tst.w d1
    bmi old_iocs_10
    ;
    ; Call the old IOCS $10 handler
    ;
    bsr old_iocs_10

modefix:
    movem.l d1/a0-a1,-(sp)
    ;
    ; Look up this mode in the timing table
    ;
    cmp.b #19,d1
    bhs skip_modefix
    ext.w d1
    lsl.w #4,d1
    lea (timing_table,pc,d1.w),a0
    ;
    ; If there's no timings for this mode, skip it
    ;
    tst.w (a0)
    beq skip_modefix
    ;
    ; Copy the new CRTC timings
    ;
    move.l #$e80000,a1
    move.w #7,d1
modefix_loop:
    move.w (a0)+,(a1)+
    dbra d1,modefix_loop
    ;
    ; Done
    ;
skip_modefix:
    movem.l (sp)+,d1/a0-a1
    rts

;
; Timing table
; All zeroes for a line means "don't change it"
;
timing_table:
    .dc.w 0,0,0,0,0,0,0,0
    .dc.w 0,0,0,0,0,0,0,0
    .dc.w 0,0,0,0,0,0,0,0
    .dc.w 0,0,0,0,0,0,0,0
    .dc.w 0,0,0,0,0,0,0,0
    .dc.w 0,0,0,0,0,0,0,0
    .dc.w 0,0,0,0,0,0,0,0
    .dc.w 0,0,0,0,0,0,0,0
    .dc.w 0,0,0,0,0,0,0,0
    .dc.w 0,0,0,0,0,0,0,0
    .dc.w 0,0,0,0,0,0,0,0
    .dc.w 0,0,0,0,0,0,0,0
    .dc.w 0,0,0,0,0,0,0,0
    .dc.w 0,0,0,0,0,0,0,0
    .dc.w 0,0,0,0,0,0,0,0
    .dc.w 0,0,0,0,0,0,0,0
    .dc.w 0,0,0,0,0,0,0,0
    .dc.w 0,0,0,0,0,0,0,0
    .dc.w 0,0,0,0,0,0,0,0

;---------------------------------------------------------------------------------------------
;
; End of resident portion
;
    .even
resident_end:

;---------------------------------------------------------------------------------------------
;
; Get next character in the file
;
getc_read:
    peaw buffer_end-buffer ; length
    lea (buffer,pc),a4
    pea (a4)               ; buffer
    move.w d7,-(sp)        ; file handle
    .dc.w $ff3f ; _READ
    lea (10,sp),sp

    tst.l d0
    ble getc_eof

    lea (a4,d0.w),a5

getc:
    cmp.l a5,a4
    bhs getc_read
    clr.l d0
    move.b (a4)+,d0
    tst.l d0
    rts

getc_eof:
    moveq.l #-1,d0
    rts

;---------------------------------------------------------------------------------------------
;
; Parse the next number from the file
;
get_number:
    move.l d2,-(sp)

get_number_nondigit_find:
    bsr getc
    bmi get_number_eof_find
    cmp.w #';',d0
    beq get_number_comment_find
    sub.w #'0',d0
    cmp.w #10,d0
    bhs get_number_nondigit_find
    move.w d0,d2

get_number_loop:
    bsr getc
    bmi get_number_eof_done
    cmp.w #';',d0
    beq get_number_comment_done
    sub.w #'0',d0
    cmp.w #10,d0
    bhs get_number_nondigit_done

    mulu.w #10,d2
    add.w d0,d2

    bra get_number_loop

get_number_comment_find:
    bsr getc
    bmi get_number_eof_find
    cmp.w #13,d0
    beq get_number_nondigit_find
    cmp.w #10,d0
    beq get_number_nondigit_find
    bra get_number_comment_find

get_number_comment_done:
    bsr getc
    bmi get_number_eof_done
    cmp.w #13,d0
    beq get_number_nondigit_done
    cmp.w #10,d0
    beq get_number_nondigit_done
    bra get_number_comment_done

get_number_eof_done:
get_number_nondigit_done:
    move.w d2,d0
    move.l (sp)+,d2
    and.l #$ffff,d0
    rts

get_number_eof_find:
    move.l (sp)+,d2
    moveq.l #-1,d0
    rts

;---------------------------------------------------------------------------------------------
;
; Print decimal number in d0.w
;
print_number:
    move.l sp,a0
    subq.l #6,sp
    clr.b -(a0) ; null terminator
print_number_loop:
    and.l #$ffff,d0
    divu.w #10,d0
    swap d0
    add.w #'0',d0
    move.b d0,-(a0)
    swap d0
    tst.w d0
    bne print_number_loop
    pea (a0)
    .dc.w $ff09 ; _PRINT
    lea (10,sp),sp
    rts

;---------------------------------------------------------------------------------------------
;
; Main (TSR setup)
;
;
; d5 = pending line?
; d6 = current mode
; d7 = file handle
;
; a2 = filename
;
; a4 = pointer to buffer
; a5 = pointer to end of buffer
;
;
    .even
main:
    lea (stacktop,pc),sp

    clr.b d5 ; pending line

    ;
    ; Display the banner
    ;
    pea (banner,pc)
    .dc.w $ff09 ; _PRINT
    addq.l #4,sp

    ;
    ; a2 (Human68k command line) ==> filename
    ;
    addq.l #1,a2 ; skip length byte

    ;
    ; If no filename was given, display usage info
    ;
    tst.b (a2)
    lea (error_usage,pc),a0
    beq error

    ;
    ; "Reading..."
    ;
    pea (reading_config,pc)
    .dc.w $ff09 ; _PRINT
    move.l a2,(sp)
    .dc.w $ff09 ; _PRINT
    addq.l #4,sp
    pea (crlf,pc)
    .dc.w $ff09 ; _PRINT
    addq.l #4,sp

    ;
    ; Open the config file for reading
    ;
    clr.w -(sp) ; MODE=0
    pea (a2)    ; filename
    .dc.w $ff3d ; _OPEN
    addq.l #6,sp

    move.w d0,d7 ; file handle
    lea (error_opening,pc),a0
    bmi error

    clr.l a5 ; end of buffer

;
; Configuration file parsing loop
;
config_loop:
    ;
    ; Get a mode number
    ;
    bsr get_number
    bmi config_eof
    move.w d0,d6

    cmp.w #19,d6
    lea (mode_out_of_range,pc),a0
    bhs mode_error

    lea (timing_table,pc),a6
    lsl.w #4,d0
    adda.w d0,a6

    tst.w (a6)
    lea (mode_specified_twice,pc),a0
    bne mode_error

    ;
    ; Copy timing data
    ;
    move.w #7,d2
copy_timing_loop:
    bsr get_number
    lea (mode_line_incomplete,pc),a0
    bmi mode_error

    move.w d0,(a6)+
    dbra d2,copy_timing_loop

    ;
    ; Verify timing data
    ;
    tst.w (-16,a6)
    lea (mode_htotal_zero,pc),a0
    beq mode_error

    tst.w (-8,a6)
    lea (mode_vtotal_zero,pc),a0
    beq mode_error

    ;
    ; List this mode as having been configured
    ;
    lea (list_first,pc),a0
    tst.b d5 ; pending line
    beq list_f
    lea (list_second,pc),a0
list_f:
    pea (a0)
    .dc.w $ff09 ; _PRINT
    addq.l #4,sp

    move.w d6,d0
    bsr print_number

    moveq.l #1,d5 ; pending line

    bra config_loop

;
; EOF encountered
;
config_eof:
    ;
    ; If there's no pending line, then no modes were specified
    ;
    tst.b d5
    lea (error_no_modes,pc),a0
    beq error

    bsr clear_pending_line

    ;
    ; Close the file
    ;
    move.w d7,-(sp) ; file handle
    .dc.w $ff3e ; _CLOSE
    addq.l #2,sp

    ;
    ; "Installing..."
    ;
    pea (installing,pc)
    .dc.w $ff09 ; _PRINT
    addq.l #4,sp

    ;
    ; Hook the IOCS handler
    ;
    pea.l (new_iocs_10,pc)
    move.w #$0110,-(sp)
    .dc.w $ff25 ; _INTVCS
    ;
    ; Store the old address
    ;
    lea (old_iocs_10_address,pc),a0
    move.l d0,(a0)
    addq.l #6,sp

    ;
    ; Get the current screen mode
    ;
    moveq.l #-1,d1
    moveq #$10,d0
    trap #15
    move.w d0,-(sp)

    ;
    ; Switch to supervisor mode
    ;
    clr.l -(sp)
    .dc.w $ff20 ; _SUPER
    move.l d0,(sp)

    ;
    ; Call modefix to set the CRTC timings directly
    ;
    move.w (4,sp),d1
    bsr modefix

    ;
    ; Switch back to user mode
    ;
    .dc.w $ff20 ; _SUPER

    ;
    ; Terminate and stay resident
    ;
    clr.w -(sp)
    peaw resident_end-resident_start
    .dc.w $ff31 ; _KEEPPR

;---------------------------------------------------------------------------------------------

attention:
    bsr clear_pending_line
    bra clear_pending_line_1

clear_pending_line:
    tst.b d5 ; pending line
    beq clear_pending_line_done
    clr.b d5
clear_pending_line_1:
    pea (crlf,pc)
    .dc.w $ff09 ; _PRINT
    addq.l #4,sp
clear_pending_line_done:
    rts

;---------------------------------------------------------------------------------------------

error:
    pea (a0)
    bsr attention
    .dc.w $ff09 ; _PRINT
    .dc.w $ff00 ; _EXIT

mode_error:
    pea (a0)
    bsr attention
    .dc.w $ff09 ; _PRINT
    move.w d6,d0
    bsr print_number
    pea (crlfcrlf,pc)
    .dc.w $ff09 ; _PRINT
    .dc.w $ff00 ; _EXIT

;---------------------------------------------------------------------------------------------

banner:
    .dc.b 'LCD flat panel mode fix for X68000 by Neill Corlett',13,10,0

error_usage:
    .dc.b 'Usage: lcdfix <configfile>'
crlfcrlf:
    .dc.b 13,10
crlf:
    .dc.b 13,10,0
reading_config:
    .dc.b 'Reading configuration file: ',0

error_opening:
    .dc.b 'Error: Unable to open configuration file',13,10,13,10,0

list_first:
    .dc.b 'Updating CRTC table for modes: ',0
list_second:
    .dc.b ',',0

mode_out_of_range:
    .dc.b 'Error: Mode number out of range: ',0
mode_specified_twice:
    .dc.b 'Error: Mode specified twice: ',0
mode_line_incomplete:
    .dc.b 'Error: Mode line incomplete: ',0
mode_htotal_zero:
    .dc.b 'Error: Horizontal total cannot be zero for mode ',0
mode_vtotal_zero:
    .dc.b 'Error: Vertical total cannot be zero for mode ',0

error_no_modes:
    .dc.b 'Error: No modes were specified in the config file',13,10,13,10,0

installing:
    .dc.b 'Installing IOCS hook',13,10,13,10,0

;---------------------------------------------------------------------------------------------

    .even
    .bss
    .even

buffer:
    .ds.b $0200
buffer_end:

;---------------------------------------------------------------------------------------------

    .even
    .stack
    .even
    .ds.b 4096
stacktop:

;---------------------------------------------------------------------------------------------
    .end main
;---------------------------------------------------------------------------------------------
